/*
 * atkdebugger.c --
 *      This is an Tcl extension to provide extended debugging futures.
 *      This extension was written for Tcl8.4.1 and require special patch
 *
 * Copyright (c) 2002 Artur Trzewik
 *
 * This is GNU GPL Lincensed Software
 *
 # This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <tcl.h>
#include <tclCompile.h>
#include <tclInt.h>

#ifndef VERSION
#define VERSION "0.20"
#endif

#define DBGNS "atk::"

#ifndef atkdebugger
endproc
#endif

typedef struct BreakpointHandle {
  int isActive;
  int sourcePosition;
  Tcl_Obj *procName;
  Tcl_Obj *expression;
  Proc *proc;
  char *breakPc;
  int counter;
} BreakpointHandle;

extern char *GetSrcInfoForPc _ANSI_ARGS_((unsigned char *pc,
        		    ByteCode* codePtr, int *lengthPtr));


static int Halt_Cmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
		int objc, Tcl_Obj *CONST objv[]));

static int
Halt_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  DebugData *ddata = (DebugData *)cdata;
  if (ddata->isActive)
    ddata->debugMode = stepInto;
  return TCL_OK;
}
static int
check_Debugcontext(DebugData *ddata, Tcl_Interp * interp)
{
  Interp *iPtr = (Interp *)interp;
  CallFrame *frame = iPtr->varFramePtr;
  if (ddata->debugLevel!=-1) {
    while (frame != ddata->debugFrame && frame!=NULL) {
      frame = frame->callerVarPtr;
    }
    // Debug Frame can also be null
    return TCL_OK;
  }
  Tcl_SetStringObj(Tcl_GetObjResult(interp),"command can be used only inside debug script",-1);
  return TCL_ERROR;
}
static int
stepInto_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  DebugData *ddata = (DebugData *)cdata;
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "stepInto");
    return TCL_ERROR;
  }
  if (check_Debugcontext(ddata,interp)==TCL_ERROR) return TCL_ERROR;
  ddata->debugMode = stepInto;
  return TCL_OK;
}
static int
stepOver_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  DebugData *ddata = (DebugData *)cdata;
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "stepOver");
    return TCL_ERROR;
  }
  if (check_Debugcontext(ddata,interp)==TCL_ERROR) return TCL_ERROR;
  ddata->debugMode = stepOver;
  return TCL_OK;
}
static int
WasError_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  DebugData *ddata = (DebugData *)cdata;
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "wasError");
    return TCL_ERROR;
  }
  if (check_Debugcontext(ddata,interp)==TCL_ERROR) return TCL_ERROR;
  Tcl_SetIntObj(Tcl_GetObjResult(interp),ddata->wasError);
  return TCL_OK;
}

static int
Debugproc_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  DebugData *ddata = (DebugData *)cdata;
  Interp *iPtr = (Interp *)interp;

  if (objc > 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "?script?");
    return TCL_ERROR;
  }
  if (objc==1) {
    if (ddata->debugproc!=NULL) {
      Tcl_SetObjResult(interp, ddata->debugproc);      
    }
  } else {
    if (ddata->debugproc!=NULL) {
      Tcl_DecrRefCount(ddata->debugproc);
    }
    Tcl_IncrRefCount(objv[1]);
    ddata->debugproc = objv[1];
  }
  return TCL_OK;
}
static int
Debuglevel_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  DebugData *ddata = (DebugData *)cdata;
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
  Tcl_SetIntObj(Tcl_GetObjResult(interp),ddata->debugLevel);  
  return TCL_OK;
}
static int
Status_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  DebugData *ddata = (DebugData *)cdata;
  if (objc > 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "?boolean?");
    return TCL_ERROR;
  }
  if (objc==1) {
    if (ddata->debugproc!=NULL) {
      Tcl_SetIntObj(Tcl_GetObjResult(interp),ddata->isActive);
    }
    return TCL_OK;
  }
  return Tcl_GetBooleanFromObj(interp, objv[1], &ddata->isActive);
}
static int
debugScript_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  DebugData *ddata = (DebugData *)cdata;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "script");
    return TCL_ERROR;
  } 
  if (ddata->isActive)
    ddata->debugMode = stepInto;
  return Tcl_EvalObjEx(interp,objv[1],TCL_EVAL_DIRECT | TCL_EVAL_GLOBAL);
}
static int
DebugByError_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  DebugData *ddata = (DebugData *)cdata;
  if (objc > 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "?boolean?");
    return TCL_ERROR;
  }
  if (objc==1) {
    if (ddata->debugproc!=NULL) {
      Tcl_SetIntObj(Tcl_GetObjResult(interp),ddata->debugByError);
    }
    return TCL_OK;
  }
  return Tcl_GetBooleanFromObj(interp, objv[1], &ddata->debugByError);
}
static int 
Breakpoint_Add(DebugData *ddata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  Tcl_CmdInfo cinfo;
  char buffer[TCL_DOUBLE_SPACE+1+5];
  int newflag,len;
  Tcl_HashEntry *entryPtr;
  BreakpointHandle *handle;
  Proc *procPtr;
  char *name;

  if (objc<3 || objc>5) {
    Tcl_WrongNumArgs(interp, 1, objv, "add procname ?position? ?expresion?");
    return TCL_ERROR;
  }
  name = Tcl_GetString(objv[2]);
  procPtr = TclFindProc((Interp *)interp, name);
  if (procPtr == NULL) {
    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		  "\"", name, "\" isn't a procedure", (char *) NULL);
    return TCL_ERROR;
  }
  len=sprintf(buffer, "break%d",ddata->breakpointsCounter++);
  Tcl_SetStringObj(Tcl_GetObjResult(interp),buffer,len);
  handle=(BreakpointHandle *)Tcl_Alloc(sizeof(BreakpointHandle));
  handle->isActive=1;
  handle->counter=0;
  handle->proc=procPtr;
  handle->breakPc=NULL;
  if (objc>3) {
    if (Tcl_GetIntFromObj(interp,objv[3],&handle->sourcePosition)!=TCL_OK) {
      Tcl_Free((char *)handle);
      return TCL_ERROR;
    }
  } else {
    handle->sourcePosition = 0;
  }
  handle->procName = objv[2];
  Tcl_IncrRefCount(objv[2]);
  if (objc>4) {
    handle->expression = objv[4];
    Tcl_IncrRefCount(objv[4]);
  } else {
    handle->expression = NULL;
  }
  entryPtr=Tcl_CreateHashEntry(&ddata->breakpointsTable,buffer,&newflag);
  Tcl_SetHashValue(entryPtr,handle);
  
  return TCL_OK;
}
BreakpointHandle *
getBreakpoint(DebugData *ddata,Tcl_Interp * interp,Tcl_Obj *nameObj)
{
  char *name = Tcl_GetStringFromObj(nameObj, NULL);  
  Tcl_HashEntry *entryPtr = Tcl_FindHashEntry(&ddata->breakpointsTable,name);

  if (entryPtr == NULL) {
    Tcl_SetStringObj(Tcl_GetObjResult(interp),"It is no breakpoint handle",-1);
    return NULL;
  } else {
    return (BreakpointHandle *)Tcl_GetHashValue(entryPtr);
  }
}

static int
Breakpoint_Info(DebugData *ddata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  Tcl_Obj *res;
  BreakpointHandle *breakpointHandle;

  if (objc>3) {
    Tcl_WrongNumArgs(interp, 1, objv, "?breakpointHandle?");
    return TCL_ERROR;
  }
 res = Tcl_GetObjResult(interp);
 if (objc==2) {
     for (entryPtr=Tcl_FirstHashEntry(&ddata->breakpointsTable,&search); 
	 entryPtr!=NULL;
	 entryPtr=Tcl_NextHashEntry(&search)) {
      Tcl_ListObjAppendElement (interp, res, Tcl_NewStringObj(Tcl_GetHashKey(&ddata->breakpointsTable,entryPtr),-1));
    }
    return TCL_OK;
  }
  if ((breakpointHandle = getBreakpoint(ddata,interp,objv[2]))==NULL) return TCL_ERROR;
  Tcl_ListObjAppendElement (interp, res, breakpointHandle->procName);
  Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(breakpointHandle->sourcePosition));
  Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(breakpointHandle->isActive));
  Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(breakpointHandle->counter));
  if (breakpointHandle->expression)
    Tcl_ListObjAppendElement (interp, res, breakpointHandle->expression);
  return TCL_OK;
}
static int
Breakpoint_Configure(DebugData *ddata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  BreakpointHandle *bhandle;
  int i,index;
  static CONST char *options[] = {
    "-active" , "-expression" , NULL
  };
  enum option {
    B_ACTIVE, B_EXPRESSION
  };
  if (objc<5 && (objc & 1)!=1) {
    Tcl_WrongNumArgs(interp, 1, objv, "breakpointHandle option value ?option value?");
    return TCL_ERROR;
  }
  if ((bhandle = getBreakpoint(ddata,interp,objv[2]))==NULL) return TCL_ERROR;
  for (i=3;i<objc;i+=2) {
    if (Tcl_GetIndexFromObj(interp, objv[i], options, "switch",
			    TCL_EXACT, &index) != TCL_OK) {
      return TCL_ERROR;
    }  
    switch (index) {
    case B_ACTIVE:
      if (Tcl_GetBooleanFromObj(interp, objv[i+1], &bhandle->isActive)==TCL_ERROR) return TCL_ERROR;
      break;
    case B_EXPRESSION:
      if (bhandle->expression) {
	Tcl_DecrRefCount(bhandle->expression);
      }
      if (Tcl_GetCharLength(objv[i+1])==0) {
	bhandle->expression = NULL;
      } else {
	bhandle->expression = objv[i+1];
	Tcl_IncrRefCount(objv[i+1]);
      }
      break;
    }
  }
  return TCL_OK;
}
static int
Breakpoint_Remove(DebugData *ddata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  Tcl_HashEntry *entryPtr;
  BreakpointHandle *bhandle;
  char *name;
  if (objc!=3) {
    Tcl_WrongNumArgs(interp, 1, objv, "breakpointHandle");
    return TCL_ERROR;
  }
  name = Tcl_GetStringFromObj(objv[2], NULL);
  entryPtr = Tcl_FindHashEntry(&ddata->breakpointsTable,name);
  if (entryPtr == NULL) {
    Tcl_SetStringObj(Tcl_GetObjResult(interp),"It is no breakpoint handle",-1);
    return TCL_ERROR;
  } else {
    bhandle = (BreakpointHandle *)Tcl_GetHashValue(entryPtr);
    Tcl_DecrRefCount(bhandle->procName);
    if (bhandle->expression) {
      Tcl_DecrRefCount(bhandle->expression);
    }
    Tcl_DeleteHashEntry(entryPtr);
  }  
  return TCL_OK;
}
static int
Breakpoint_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  int index,result;
  static CONST char *options[] = {
    "add" , "info" , "configure" , "remove", NULL
  };
  enum option {
    B_ADD, B_INFO, B_CONFIGURE , B_REMOVE
  };
  if (objc <= 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "subcommand ??");
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp, objv[1], options, "switch",
                TCL_EXACT, &index) != TCL_OK) {
      return TCL_ERROR;
  }
  result = TCL_ERROR;
  switch (index) {
  case B_ADD:
    result = Breakpoint_Add(cdata, interp, objc, objv);
    break;
  case B_INFO:
    result = Breakpoint_Info(cdata, interp, objc, objv);
    break;
  case B_CONFIGURE:
    result = Breakpoint_Configure(cdata, interp, objc, objv);
    break;
  case B_REMOVE:
    result = Breakpoint_Remove(cdata, interp, objc, objv);
    break;
  }
  return result;
}
/*
  Try to compute script and the evaluation place of it
  It must handle 4 cases (2 * 2)
  global or frame context
  bytecoded or direct evaluation
 */
static int
Runplace_Cmd(ClientData cdata, Tcl_Interp * interp, int objc, Tcl_Obj *CONST objv[])
{
  DebugData *ddata = (DebugData *)cdata;
  Interp *iPtr = (Interp *)interp;
  int level,length,result;
  Tcl_Obj *res;
  ByteCode *codePtr;
  CallFrame *frame;
  char *byte,*pc;
  char levelDescription[20];

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "level");
    return TCL_ERROR;
  }
  if (Tcl_GetIntFromObj(interp, objv[1], &level) != TCL_OK) return TCL_ERROR;
  if (check_Debugcontext(ddata,interp)==TCL_ERROR) return TCL_ERROR;

  if (level!=0) {
    sprintf(levelDescription,"#%d",level);
    result = TclGetFrame(interp,levelDescription,&frame);
    
    if (result==-1) return TCL_ERROR;
  } else {
    frame = NULL;
  }
  res = Tcl_GetObjResult(interp);

  if (frame==NULL) {
    if (ddata->script!=NULL) {
      Tcl_ListObjAppendElement (interp, res, Tcl_NewStringObj(ddata->script,-1));
      Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(ddata->command-ddata->script));
      Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(ddata->length));
    } else {
       pc = ddata->debugpc;
       codePtr = (ByteCode*)ddata->codePtr;
       byte = GetSrcInfoForPc(pc, codePtr, &length);
       Tcl_ListObjAppendElement (interp, res, Tcl_NewStringObj(codePtr->source,-1));
       Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(byte-codePtr->source));
       Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(length));   
    }
    return TCL_OK;
  }

  if (frame->script!=NULL) {
    Tcl_ListObjAppendElement (interp, res, Tcl_NewStringObj(frame->script,-1));
    Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(frame->command-frame->script));
    Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(frame->length));
    return TCL_OK;
  }
  
  codePtr = (ByteCode *)frame->codePtr;
  if (codePtr==NULL) {
    Tcl_ListObjAppendElement (interp, res, Tcl_NewStringObj("can not compute run place",-1));
    Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(0));
    Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(3));
    return TCL_OK;
  }
  
  if (level==ddata->debugLevel) {
    pc = ddata->debugpc;
  } else {
    pc = frame->pc;
  }
  byte = GetSrcInfoForPc(pc, codePtr, &length);
  res = Tcl_GetObjResult(interp);
  Tcl_ListObjAppendElement (interp, res, Tcl_NewStringObj(codePtr->source,-1));
  Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(byte-codePtr->source));
  Tcl_ListObjAppendElement (interp, res, Tcl_NewIntObj(length));
  return TCL_OK;
}
static int
debugEntryProc(Tcl_Interp *interp,int wasError) {
  DebugData *ddata = ((Interp *)interp)->debugStruct;
  Interp *iPtr = (Interp *)interp;
  CallFrame *frame = iPtr->varFramePtr;
  Tcl_Obj *resultSave;

  ddata->wasError = wasError;
  if (wasError) {
    while (frame!=NULL) {
      if (*frame->catchTop!=-1) {
	return TCL_OK;
      }
      frame = frame->callerVarPtr;
    }
  }
  if (iPtr->varFramePtr==NULL) {
    ddata->debugLevel = 0;
  } else {
    ddata->debugLevel = iPtr->varFramePtr->level;
  }
  ddata->debugFrame = iPtr->varFramePtr;
  ddata->debugMode = none;
  if (iPtr->varFramePtr!=NULL)
    ddata->debugpc = iPtr->varFramePtr->pc;
  resultSave = Tcl_GetObjResult(interp);
  Tcl_IncrRefCount(resultSave);
  Tcl_EvalObjEx(interp,ddata->debugproc,0);
  Tcl_SetObjResult(interp,resultSave);
  Tcl_DecrRefCount(resultSave);
  ddata->debugLevel = -1;
  return TCL_OK;
}
/* Entry Point for callback from Tcl internals
   assumption: Only one breakpoint pro method
   The Procedure is entered than the method is checked
   if it contain breakpoint. If yes the CallFrame store the pointer to breakPoinHandle
   It will be checkecd von every breakpoint loop
*/
static int
checkBreakpointsProc(Tcl_Interp *interp,int isProcEntryMode) {
  DebugData *ddata = ((Interp *)interp)->debugStruct;
  Interp *iPtr = (Interp *)interp;
  Proc *proc;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  BreakpointHandle *handle;
  CallFrame *frame = iPtr->varFramePtr;
  int expressionResult;

  if (isProcEntryMode) {
    frame->breakpointHandle = NULL;
    for (entryPtr=Tcl_FirstHashEntry(&ddata->breakpointsTable,&search); 
	 entryPtr!=NULL;
	 entryPtr=Tcl_NextHashEntry(&search)) {
      handle = (BreakpointHandle *)Tcl_GetHashValue(entryPtr);
      if (frame->procPtr==handle->proc) {
	if (handle->isActive)
	  frame->breakpointHandle = (void *)handle;
	break;
      }
    }
  } else {
    handle = (BreakpointHandle *)frame->breakpointHandle;
    handle->counter++;
    if (checkBreakPoint(iPtr,handle)) {
      if (handle->expression!=NULL) {
	ddata->debugLevel = 0;
	Tcl_EvalObjEx(interp,handle->expression,0);
	ddata->debugLevel = -1;
	if (Tcl_GetBooleanFromObj(interp,Tcl_GetObjResult(interp),&expressionResult)!=TCL_ERROR) {
	  if (expressionResult) ddata->debugMode=stepInto; 
	}
      } else {
	ddata->debugMode=stepInto;
      }
      frame->breakpointHandle = NULL;
    }
  }
  
}
int checkBreakPoint(Interp *iPtr,BreakpointHandle *handle) {
  CallFrame *frame;
  char *byte,*pc;
  int length;
  ByteCode *codePtr;

  frame = iPtr->varFramePtr;
  if (handle->sourcePosition==0) return 1;
  pc = frame->pc;

  if (handle->breakPc==NULL) {
    codePtr = (ByteCode *)frame->procPtr->bodyPtr->internalRep.otherValuePtr;
    byte = GetSrcInfoForPc(pc, codePtr, &length);
    if (byte-codePtr->source > handle->sourcePosition) {
      handle->breakPc = pc;
      return 1;
    } else {
      return 0;
    }
  } else {
    return pc==handle->breakPc;
  }
}
/*
 *----------------------------------------------------------------------
 *
 * Sample_Init --
 *
 *	Initialize the new package.  The string "Sample" in the
 *	function name must match the PACKAGE declaration at the top of
 *	configure.in.
 *
 * Results:
 *	A standard Tcl result
 *
 * Side effects:
 *	The Tclsha1 package is created.
 *	One new command "sha1" is added to the Tcl interpreter.
 *
 *----------------------------------------------------------------------
 */

int Atkdebugger_Init(Tcl_Interp *interp)
{
    /*
     * This may work with 8.0, but we are using strictly stubs here,
     * which requires 8.1.
     */
  DebugData *ddata = (DebugData *)Tcl_Alloc(sizeof(DebugData));
  ddata->debugproc = NULL;
  ddata->debugLevel = -1;
  ddata->debugEntryProc = debugEntryProc;
  ddata->checkBreakpointsProc = checkBreakpointsProc;
  ddata->debugMode = none;
  ddata->debugFrame = NULL;
  ddata->isActive = 1;
  ddata->debugByError = 1;
  ddata->breakpointsCounter = 0;
  ddata->wasError = 0;

  Tcl_InitHashTable (&ddata->breakpointsTable, TCL_STRING_KEYS);

  ((Interp *)interp)->debugStruct = ddata;

  if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
    return TCL_ERROR;
  }
  if (Tcl_PkgRequire(interp, "Tcl", "8.1", 0) == NULL) {
    return TCL_ERROR;
  }
  if (Tcl_PkgProvide(interp, "atkdebugger", VERSION) != TCL_OK) {
    return TCL_ERROR;
  }
  Tcl_CreateObjCommand(interp, DBGNS"halt", Halt_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, DBGNS"stepInto", stepInto_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, DBGNS"stepOver", stepOver_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, DBGNS"debugproc", Debugproc_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, DBGNS"debuglevel", Debuglevel_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, DBGNS"runplace", Runplace_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, DBGNS"debugstatus", Status_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, DBGNS"debugByError", DebugByError_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, DBGNS"wasError", WasError_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, DBGNS"breakpoint", Breakpoint_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, DBGNS"debugScript", debugScript_Cmd,
		    (ClientData)ddata, (Tcl_CmdDeleteProc *)NULL);
  return TCL_OK;
}
